@@ -15,86 +15,100 @@ ESCAPE_REGEX = re.compile(r"[\0\n\r\032\'\"\\]")
 ESCAPE_MAP = {'\0': '\\0', '\n': '\\n', '\r': '\\r', '\032': '\\Z',
               '\'': '\\\'', '"': '\\"', '\\': '\\\\'}
 
+def escape_item(val, charset, mapping=None):
+    if mapping is None:
+        mapping = encoders
+    encoder = mapping.get(type(val))
 
-def escape_item(val, charset):
-    if type(val) in [tuple, list, set]:
-        return escape_sequence(val, charset)
-    if type(val) is dict:
-        return escape_dict(val, charset)
-    encoder = encoders[type(val)]
-    val = encoder(val)
-    if type(val) in [str, int, text_type]:
-        return val
+    # Fallback to default when no encoder found
+    if not encoder:
+        try:
+            encoder = mapping[text_type]
+        except KeyError:
+            raise TypeError("no default type converter defined")
+
+    if encoder in (escape_dict, escape_sequence):
+        val = encoder(val, charset, mapping)
+    else:
+        val = encoder(val, mapping)
     return val
 
-def escape_dict(val, charset):
+def escape_dict(val, charset, mapping=None):
     n = {}
     for k, v in val.items():
-        quoted = escape_item(v, charset)
+        quoted = escape_item(v, charset, mapping)
         n[k] = quoted
     return n
 
-def escape_sequence(val, charset):
+def escape_sequence(val, charset, mapping=None):
     n = []
     for item in val:
-        quoted = escape_item(item, charset)
+        quoted = escape_item(item, charset, mapping)
         n.append(quoted)
     return "(" + ",".join(n) + ")"
 
-def escape_set(val, charset):
-    val = map(lambda x: escape_item(x, charset), val)
+def escape_set(val, charset, mapping=None):
+    val = map(lambda x: escape_item(x, charset, mapping), val)
     return ','.join(val)
 
-def escape_bool(value):
+def escape_bool(value, mapping=None):
     return str(int(value))
 
-def escape_object(value):
+def escape_object(value, mapping=None):
     return str(value)
 
-def escape_int(value):
+def escape_int(value, mapping=None):
     return str(value)
 
-
-def escape_float(value):
+def escape_float(value, mapping=None):
     return ('%.15g' % value)
 
-def escape_string(value):
+def escape_string(value, mapping=None):
     return ("%s" % (ESCAPE_REGEX.sub(
             lambda match: ESCAPE_MAP.get(match.group(0)), value),))
 
-def escape_str(value):
-    return "'%s'" % escape_string(value)
+def escape_str(value, mapping=None):
+    return "'%s'" % escape_string(value, mapping)
 
-def escape_unicode(value):
-    return escape_str(value)
+def escape_unicode(value, mapping=None):
+    return escape_str(value, mapping)
 
-def escape_bytes(value):
-    return "x'%s'" % binascii.hexlify(value).decode(sys.getdefaultencoding())
+def escape_bytes(value, mapping=None):
+    # escape_bytes is calld only on Python 3.
+    return escape_str(value.decode('ascii', 'surrogateescape'), mapping)
 
-def escape_None(value):
+def escape_None(value, mapping=None):
     return 'NULL'
 
-def escape_timedelta(obj):
+def escape_timedelta(obj, mapping=None):
     seconds = int(obj.seconds) % 60
     minutes = int(obj.seconds // 60) % 60
     hours = int(obj.seconds // 3600) % 24 + int(obj.days) * 24
-    return escape_str('%02d:%02d:%02d' % (hours, minutes, seconds))
+    if obj.microseconds:
+        fmt = "'{0:02d}:{1:02d}:{2:02d}.{3:06d}'"
+    else:
+        fmt = "'{0:02d}:{1:02d}:{2:02d}'"
+    return fmt.format(hours, minutes, seconds, obj.microseconds)
 
-def escape_time(obj):
-    s = "%02d:%02d:%02d" % (int(obj.hour), int(obj.minute),
-                            int(obj.second))
+def escape_time(obj, mapping=None):
     if obj.microsecond:
-        s += ".{0:06}".format(obj.microsecond)
+        fmt = "'{0.hour:02}:{0.minute:02}:{0.second:02}.{0.microsecond:06}'"
+    else:
+        fmt = "'{0.hour:02}:{0.minute:02}:{0.second:02}'"
+    return fmt.format(obj)
 
-    return escape_str(s)
+def escape_datetime(obj, mapping=None):
+    if obj.microsecond:
+        fmt = "'{0.year:04}-{0.month:02}-{0.day:02} {0.hour:02}:{0.minute:02}:{0.second:02}.{0.microsecond:06}'"
+    else:
+        fmt = "'{0.year:04}-{0.month:02}-{0.day:02} {0.hour:02}:{0.minute:02}:{0.second:02}'"
+    return fmt.format(obj)
 
-def escape_datetime(obj):
-    return escape_str(obj.isoformat(' '))
+def escape_date(obj, mapping=None):
+    fmt = "'{0.year:04}-{0.month:02}-{0.day:02}'"
+    return fmt.format(obj)
 
-def escape_date(obj):
-    return escape_str(obj.isoformat())
-
-def escape_struct_time(obj):
+def escape_struct_time(obj, mapping=None):
     return escape_datetime(datetime.datetime(*obj[:6]))
 
 def convert_datetime(obj):
@@ -154,12 +168,16 @@ def convert_timedelta(obj):
             (obj, tail) = obj.split('.')
             microseconds = float('0.' + tail) * 1e6
         hours, minutes, seconds = obj.split(':')
+        negate = 1
+        if hours.startswith("-"):
+            hours = hours[1:]
+            negate = -1
         tdelta = datetime.timedelta(
             hours = int(hours),
             minutes = int(minutes),
             seconds = int(seconds),
             microseconds = int(microseconds)
-            )
+            ) * negate
         return tdelta
     except ValueError:
         return None
@@ -252,13 +270,19 @@ def convert_mysql_timestamp(timestamp):
 def convert_set(s):
     return set(s.split(","))
 
-def convert_bit(b):
-    #b = "\x00" * (8 - len(b)) + b # pad w/ zeroes
-    #return struct.unpack(">Q", b)[0]
-    #
-    # the snippet above is right, but MySQLdb doesn't process bits,
-    # so we shouldn't either
-    return b
+
+def through(x):
+    return x
+
+
+#def convert_bit(b):
+#    b = "\x00" * (8 - len(b)) + b # pad w/ zeroes
+#    return struct.unpack(">Q", b)[0]
+#    
+#     the snippet above is right, but MySQLdb doesn't process bits,
+#     so we shouldn't either
+convert_bit = through
+
 
 def convert_characters(connection, field, data):
     field_charset = charset_by_id(field.charsetnr).name
@@ -276,59 +300,53 @@ def convert_characters(connection, field, data):
     return data
 
 encoders = {
-        bool: escape_bool,
-        int: escape_int,
-        long_type: escape_int,
-        float: escape_float,
-        str: escape_str,
-        text_type: escape_unicode,
-        tuple: escape_sequence,
-        list: escape_sequence,
-        set: escape_sequence,
-        dict: escape_dict,
-        type(None): escape_None,
-        datetime.date: escape_date,
-        datetime.datetime: escape_datetime,
-        datetime.timedelta: escape_timedelta,
-        datetime.time: escape_time,
-        time.struct_time: escape_struct_time,
-        Decimal: str,
-        }
-
-
-def through(x):
-    return x
+    bool: escape_bool,
+    int: escape_int,
+    long_type: escape_int,
+    float: escape_float,
+    str: escape_str,
+    text_type: escape_unicode,
+    tuple: escape_sequence,
+    list: escape_sequence,
+    set: escape_sequence,
+    dict: escape_dict,
+    type(None): escape_None,
+    datetime.date: escape_date,
+    datetime.datetime: escape_datetime,
+    datetime.timedelta: escape_timedelta,
+    datetime.time: escape_time,
+    time.struct_time: escape_struct_time,
+    Decimal: escape_object,
+}
 
 if not PY2 or JYTHON or IRONPYTHON:
     encoders[bytes] = escape_bytes
 
 decoders = {
-        FIELD_TYPE.BIT: convert_bit,
-        FIELD_TYPE.TINY: int,
-        FIELD_TYPE.SHORT: int,
-        FIELD_TYPE.LONG: int,
-        FIELD_TYPE.FLOAT: float,
-        FIELD_TYPE.DOUBLE: float,
-        FIELD_TYPE.DECIMAL: float,
-        FIELD_TYPE.NEWDECIMAL: float,
-        FIELD_TYPE.LONGLONG: int,
-        FIELD_TYPE.INT24: int,
-        FIELD_TYPE.YEAR: int,
-        FIELD_TYPE.TIMESTAMP: convert_mysql_timestamp,
-        FIELD_TYPE.DATETIME: convert_datetime,
-        FIELD_TYPE.TIME: convert_timedelta,
-        FIELD_TYPE.DATE: convert_date,
-        FIELD_TYPE.SET: convert_set,
-        FIELD_TYPE.BLOB: through,
-        FIELD_TYPE.TINY_BLOB: through,
-        FIELD_TYPE.MEDIUM_BLOB: through,
-        FIELD_TYPE.LONG_BLOB: through,
-        FIELD_TYPE.STRING: through,
-        FIELD_TYPE.VAR_STRING: through,
-        FIELD_TYPE.VARCHAR: through,
-        FIELD_TYPE.DECIMAL: Decimal,
-        FIELD_TYPE.NEWDECIMAL: Decimal,
-        }
+    FIELD_TYPE.BIT: convert_bit,
+    FIELD_TYPE.TINY: int,
+    FIELD_TYPE.SHORT: int,
+    FIELD_TYPE.LONG: int,
+    FIELD_TYPE.FLOAT: float,
+    FIELD_TYPE.DOUBLE: float,
+    FIELD_TYPE.LONGLONG: int,
+    FIELD_TYPE.INT24: int,
+    FIELD_TYPE.YEAR: int,
+    FIELD_TYPE.TIMESTAMP: convert_mysql_timestamp,
+    FIELD_TYPE.DATETIME: convert_datetime,
+    FIELD_TYPE.TIME: convert_timedelta,
+    FIELD_TYPE.DATE: convert_date,
+    FIELD_TYPE.SET: convert_set,
+    FIELD_TYPE.BLOB: through,
+    FIELD_TYPE.TINY_BLOB: through,
+    FIELD_TYPE.MEDIUM_BLOB: through,
+    FIELD_TYPE.LONG_BLOB: through,
+    FIELD_TYPE.STRING: through,
+    FIELD_TYPE.VAR_STRING: through,
+    FIELD_TYPE.VARCHAR: through,
+    FIELD_TYPE.DECIMAL: Decimal,
+    FIELD_TYPE.NEWDECIMAL: Decimal,
+}
 
 
 # for MySQLdb compatibility
