@@ -1,19 +1,17 @@
 # -*- coding: utf-8 -*-
 from __future__ import print_function, absolute_import
 import re
+import warnings
 
 from ._compat import range_type, text_type, PY2
 
-from .err import (
-    Warning, Error, InterfaceError, DataError,
-    DatabaseError, OperationalError, IntegrityError, InternalError,
-    NotSupportedError, ProgrammingError)
+from . import err
 
 
 #: Regular expression for :meth:`Cursor.executemany`.
 #: executemany only suports simple bulk insert.
 #: You can use it to load large dataset.
-RE_INSERT_VALUES = re.compile(r"""INSERT\s.+\sVALUES\s+(\(\s*%s\s*(,\s*%s\s*)*\))\s*\Z""",
+RE_INSERT_VALUES = re.compile(r"""(INSERT\s.+\sVALUES\s+)(\(\s*%s\s*(?:,\s*%s\s*)*\))(\s*(?:ON DUPLICATE.*)?)\Z""",
                               re.IGNORECASE | re.DOTALL)
 
 
@@ -42,12 +40,6 @@ class Cursor(object):
         self._result = None
         self._rows = None
 
-    def __del__(self):
-        '''
-        When this gets GC'd close it.
-        '''
-        self.close()
-
     def close(self):
         '''
         Closing a cursor just exhausts all remaining data.
@@ -63,12 +55,12 @@ class Cursor(object):
 
     def _get_db(self):
         if not self.connection:
-            raise ProgrammingError("Cursor closed")
+            raise err.ProgrammingError("Cursor closed")
         return self.connection
 
     def _check_executed(self):
         if not self._executed:
-            raise ProgrammingError("execute() first")
+            raise err.ProgrammingError("execute() first")
 
     def _conv_row(self, row):
         return row
@@ -147,39 +139,48 @@ class Cursor(object):
 
         m = RE_INSERT_VALUES.match(query)
         if m:
-            q_values = m.group(1).rstrip()
+            q_prefix = m.group(1)
+            q_values = m.group(2).rstrip()
+            q_postfix = m.group(3) or ''
             assert q_values[0] == '(' and q_values[-1] == ')'
-            q_prefix = query[:m.start(1)]
-            return self._do_execute_many(q_prefix, q_values, args,
+            return self._do_execute_many(q_prefix, q_values, q_postfix, args,
                                          self.max_stmt_length,
                                          self._get_db().encoding)
 
         self.rowcount = sum(self.execute(query, arg) for arg in args)
         return self.rowcount
 
-    def _do_execute_many(self, prefix, values, args, max_stmt_length, encoding):
+    def _do_execute_many(self, prefix, values, postfix, args, max_stmt_length, encoding):
         conn = self._get_db()
         escape = self._escape_args
         if isinstance(prefix, text_type):
             prefix = prefix.encode(encoding)
+        if isinstance(postfix, text_type):
+            postfix = postfix.encode(encoding)
         sql = bytearray(prefix)
         args = iter(args)
         v = values % escape(next(args), conn)
         if isinstance(v, text_type):
-            v = v.encode(encoding)
+            if PY2:
+                v = v.encode(encoding)
+            else:
+                v = v.encode(encoding, 'surrogateescape')
         sql += v
         rows = 0
         for arg in args:
             v = values % escape(arg, conn)
             if isinstance(v, text_type):
-                v = v.encode(encoding)
-            if len(sql) + len(v) + 1 > max_stmt_length:
-                rows += self.execute(sql)
+                if PY2:
+                    v = v.encode(encoding)
+                else:
+                    v = v.encode(encoding, 'surrogateescape')
+            if len(sql) + len(v) + len(postfix) + 1 > max_stmt_length:
+                rows += self.execute(sql + postfix)
                 sql = bytearray(prefix)
             else:
                 sql += b','
             sql += v
-        rows += self.execute(sql)
+        rows += self.execute(sql + postfix)
         self.rowcount = rows
         return rows
 
@@ -237,7 +238,7 @@ class Cursor(object):
         ''' Fetch several rows '''
         self._check_executed()
         if self._rows is None:
-            return None
+            return ()
         end = self.rownumber + (size or self.arraysize)
         result = self._rows[self.rownumber:end]
         self.rownumber = min(end, len(self._rows))
@@ -247,7 +248,7 @@ class Cursor(object):
         ''' Fetch all the rows '''
         self._check_executed()
         if self._rows is None:
-            return None
+            return ()
         if self.rownumber:
             result = self._rows[self.rownumber:]
         else:
@@ -262,7 +263,7 @@ class Cursor(object):
         elif mode == 'absolute':
             r = value
         else:
-            raise ProgrammingError("unknown scroll mode %s" % mode)
+            raise err.ProgrammingError("unknown scroll mode %s" % mode)
 
         if not (0 <= r < len(self._rows)):
             raise IndexError("out of range")
@@ -286,19 +287,27 @@ class Cursor(object):
         self.lastrowid = result.insert_id
         self._rows = result.rows
 
+        if result.warning_count > 0:
+            self._show_warnings(conn)
+
+    def _show_warnings(self, conn):
+        ws = conn.show_warnings()
+        for w in ws:
+            warnings.warn(w[-1], err.Warning, 4)
+
     def __iter__(self):
         return iter(self.fetchone, None)
 
-    Warning = Warning
-    Error = Error
-    InterfaceError = InterfaceError
-    DatabaseError = DatabaseError
-    DataError = DataError
-    OperationalError = OperationalError
-    IntegrityError = IntegrityError
-    InternalError = InternalError
-    ProgrammingError = ProgrammingError
-    NotSupportedError = NotSupportedError
+    Warning = err.Warning
+    Error = err.Error
+    InterfaceError = err.InterfaceError
+    DatabaseError = err.DatabaseError
+    DataError = err.DataError
+    OperationalError = err.OperationalError
+    IntegrityError = err.IntegrityError
+    InternalError = err.InternalError
+    ProgrammingError = err.ProgrammingError
+    NotSupportedError = err.NotSupportedError
 
 
 class DictCursorMixin(object):
@@ -390,7 +399,6 @@ class SSCursor(Cursor):
         Fetch all, as per MySQLdb. Pretty useless for large queries, as
         it is buffered. See fetchall_unbuffered(), if you want an unbuffered
         generator version of this method.
-
         """
         return list(self.fetchall_unbuffered())
 
@@ -407,7 +415,6 @@ class SSCursor(Cursor):
 
     def fetchmany(self, size=None):
         """ Fetch many """
-
         self._check_executed()
         if size is None:
             size = self.arraysize
@@ -426,7 +433,7 @@ class SSCursor(Cursor):
 
         if mode == 'relative':
             if value < 0:
-                raise NotSupportedError(
+                raise err.NotSupportedError(
                         "Backwards scrolling not supported by this cursor")
 
             for _ in range_type(value):
@@ -434,7 +441,7 @@ class SSCursor(Cursor):
             self.rownumber += value
         elif mode == 'absolute':
             if value < self.rownumber:
-                raise NotSupportedError(
+                raise err.NotSupportedError(
                     "Backwards scrolling not supported by this cursor")
 
             end = value - self.rownumber
@@ -442,7 +449,7 @@ class SSCursor(Cursor):
                 self.read_next()
             self.rownumber = value
         else:
-            raise ProgrammingError("unknown scroll mode %s" % mode)
+            raise err.ProgrammingError("unknown scroll mode %s" % mode)
 
 
 class SSDictCursor(DictCursorMixin, SSCursor):
