@@ -6,6 +6,10 @@ from pymysql.err import ProgrammingError
 
 import time
 import datetime
+import warnings
+
+from unittest2 import SkipTest
+
 
 __all__ = ["TestConversion", "TestCursor", "TestBulkInserts"]
 
@@ -18,16 +22,13 @@ class TestConversion(base.PyMySQLTestCase):
         c.execute("create table test_datatypes (b bit, i int, l bigint, f real, s varchar(32), u varchar(32), bb blob, d date, dt datetime, ts timestamp, td time, t time, st datetime)")
         try:
             # insert values
-            v = (True, -3, 123456789012, 5.7, "hello'\" world", u"Espa\xc3\xb1ol", "binary\x00data".encode(conn.charset), datetime.date(1988,2,2), datetime.datetime.now(), datetime.timedelta(5,6), datetime.time(16,32), time.localtime())
+
+            v = (True, -3, 123456789012, 5.7, "hello'\" world", u"Espa\xc3\xb1ol", "binary\x00data".encode(conn.charset), datetime.date(1988,2,2), datetime.datetime(2014, 5, 15, 7, 45, 57), datetime.timedelta(5,6), datetime.time(16,32), time.localtime())
             c.execute("insert into test_datatypes (b,i,l,f,s,u,bb,d,dt,td,t,st) values (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)", v)
             c.execute("select b,i,l,f,s,u,bb,d,dt,td,t,st from test_datatypes")
             r = c.fetchone()
             self.assertEqual(util.int2byte(1), r[0])
-            self.assertEqual(v[1:8], r[1:8])
-            # mysql throws away microseconds so we need to check datetimes
-            # specially. additionally times are turned into timedeltas.
-            self.assertEqual(datetime.datetime(*v[8].timetuple()[:6]), r[8])
-            self.assertEqual(v[9], r[9]) # just timedeltas
+            self.assertEqual(v[1:10], r[1:10])
             self.assertEqual(datetime.timedelta(0, 60 * (v[10].hour * 60 + v[10].minute)), r[10])
             self.assertEqual(datetime.datetime(*v[-1][:6]), r[-1])
 
@@ -122,21 +123,24 @@ class TestConversion(base.PyMySQLTestCase):
                           -datetime.timedelta(0, 1800)),
                          c.fetchone())
 
-    def test_datetime(self):
-        """ test datetime conversion """
+    def test_datetime_microseconds(self):
+        """ test datetime conversion w microseconds"""
+
         conn = self.connections[0]
+        if not self.mysql_server_is(conn, (5, 6, 4)):
+            raise SkipTest("target backend does not support microseconds")
         c = conn.cursor()
-        dt = datetime.datetime(2013,11,12,9,9,9,123450)
+        dt = datetime.datetime(2013, 11, 12, 9, 9, 9, 123450)
+        c.execute("create table test_datetime (id int, ts datetime(6))")
         try:
-            c.execute("create table test_datetime (id int, ts datetime(6))")
-            c.execute("insert into test_datetime values (1,'2013-11-12 09:09:09.12345')")
+            c.execute(
+                "insert into test_datetime values (%s, %s)",
+                (1, dt)
+            )
             c.execute("select ts from test_datetime")
-            self.assertEqual((dt,),c.fetchone())
-        except ProgrammingError:
-            # User is running a version of MySQL that doesn't support msecs within datetime
-            pass
+            self.assertEqual((dt,), c.fetchone())
         finally:
-            c.execute("drop table if exists test_datetime")
+            c.execute("drop table test_datetime")
 
 
 class TestCursor(base.PyMySQLTestCase):
@@ -223,14 +227,14 @@ class TestCursor(base.PyMySQLTestCase):
         """ test a single tuple """
         conn = self.connections[0]
         c = conn.cursor()
-        try:
-            c.execute("create table mystuff (id integer primary key)")
-            c.execute("insert into mystuff (id) values (1)")
-            c.execute("insert into mystuff (id) values (2)")
-            c.execute("select id from mystuff where id in %s", ((1,),))
-            self.assertEqual([(1,)], list(c.fetchall()))
-        finally:
-            c.execute("drop table mystuff")
+        self.safe_create_table(
+            conn, 'mystuff',
+            "create table mystuff (id integer primary key)")
+        c.execute("insert into mystuff (id) values (1)")
+        c.execute("insert into mystuff (id) values (2)")
+        c.execute("select id from mystuff where id in %s", ((1,),))
+        self.assertEqual([(1,)], list(c.fetchall()))
+        c.close()
 
 
 class TestBulkInserts(base.PyMySQLTestCase):
@@ -243,9 +247,8 @@ class TestBulkInserts(base.PyMySQLTestCase):
         c = conn.cursor(self.cursor_type)
 
         # create a table ane some data to query
-        c.execute("drop table if exists bulkinsert")
-        c.execute(
-"""CREATE TABLE bulkinsert
+        self.safe_create_table(conn, 'bulkinsert', """\
+CREATE TABLE bulkinsert
 (
 id int(11),
 name char(20),
@@ -287,7 +290,7 @@ values (%s,
 %s , %s,
 %s )
  """, data)
-        self.assertEqual(cursor._last_executed, bytearray(b"""insert
+        self.assertEqual(cursor._last_executed.strip(), bytearray(b"""insert
 into bulkinsert (id, name,
 age, height)
 values (0,
@@ -309,7 +312,39 @@ values (0,
         cursor.execute('commit')
         self._verify_records(data)
 
+    def test_issue_288(self):
+        """executemany should work with "insert ... on update" """
+        conn = self.connections[0]
+        cursor = conn.cursor()
+        data = [(0, "bob", 21, 123), (1, "jim", 56, 45), (2, "fred", 100, 180)]
+        cursor.executemany("""insert
+into bulkinsert (id, name,
+age, height)
+values (%s,
+%s , %s,
+%s ) on duplicate key update
+age = values(age)
+ """, data)
+        self.assertEqual(cursor._last_executed.strip(), bytearray(b"""insert
+into bulkinsert (id, name,
+age, height)
+values (0,
+'bob' , 21,
+123 ),(1,
+'jim' , 56,
+45 ),(2,
+'fred' , 100,
+180 ) on duplicate key update
+age = values(age)"""))
+        cursor.execute('commit')
+        self._verify_records(data)
 
-if __name__ == "__main__":
-    import unittest
-    unittest.main()
+    def test_warnings(self):
+        con = self.connections[0]
+        cur = con.cursor()
+        with warnings.catch_warnings(record=True) as ws:
+            warnings.simplefilter("always")
+            cur.execute("drop table if exists no_exists_table")
+        self.assertEqual(len(ws), 1)
+        self.assertEqual(ws[0].category, pymysql.Warning)
+        self.assertTrue(u"no_exists_table" in str(ws[0].message))
