@@ -1,11 +1,8 @@
+import unittest2
+
 from pymysql.tests import base
 from pymysql import util
 
-try:
-    import unittest2 as unittest
-except ImportError:
-    import unittest
-
 
 class TestNextset(base.PyMySQLTestCase):
 
@@ -32,7 +29,16 @@ class TestNextset(base.PyMySQLTestCase):
         cur.execute("SELECT 42")
         self.assertEqual([(42,)], list(cur))
 
-    @unittest.expectedFailure
+    def test_ok_and_next(self):
+        cur = self.con.cursor()
+        cur.execute("SELECT 1; commit; SELECT 2;")
+        self.assertEqual([(1,)], list(cur))
+        self.assertTrue(cur.nextset())
+        self.assertTrue(cur.nextset())
+        self.assertEqual([(2,)], list(cur))
+        self.assertFalse(bool(cur.nextset()))
+
+    @unittest2.expectedFailure
     def test_multi_cursor(self):
         cur1 = self.con.cursor()
         cur2 = self.con.cursor()
